class Tokenizer(object):
    """An iterator over the tokens of a Treebank."""
    def __init__(self, input):
        self.input = input
        self.line = []

    def __next__(self):
        """Return the next token."""
        while self.line == []:
            s = self.input.readline()
            if s == '': 
                raise StopIteration
            self.line = s.replace('(', ' ( ').replace(')', ' ) ').split()
        return self.line.pop(0)

    def __iter__(self):
        return self

def read_trees(input):
    """Yield trees in a file."""
    tokens = Tokenizer(input)
    while True:
        assert next(tokens) == '('
        assert next(tokens) == '('
        tree = read_tree(tokens)
        yield Tree('ROOT', [START, tree, END])
        assert next(tokens) == ')'
    
def read_tree(tokens):
    """Read the next well-formed tree."""
    tag = next(tokens).split('-')[0].split('=')[0].split('|')[0]
    element = next(tokens)
    if element != '(':
        assert next(tokens) == ')'
        return Leaf(tag, element)
    branches = []
    while element != ')':
        assert element == '('
        branches.append(read_tree(tokens))
        element = next(tokens)
    return Tree(tag, branches)

class Leaf(object):
    """A word labeled with a syntactic category."""
    def __init__(self, tag, word):
        self.tag = tag
        self.word = word

    def __str__(self):
        return '({0} {1})'.format(self.tag, self.word)

START = Leaf('<S>', '<S>')
END = Leaf('</S>', '</S>')

class Tree(object):
    """A syntactic constituent."""
    def __init__(self, tag, branches):
        self.tag = tag
        self.branches = branches

    def __str__(self):
        branch_strings = ' '.join(str(b) for b in self.branches)
        return '({0} {1})'.format(self.tag, branch_strings)
